#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <windows.h>
#include <winnt.h>
#include <vector>
#include "PE.h"

using namespace std;

PESection::PESection()
{
  this->Data = NULL;
  memset(&this->Header, 0, sizeof(this->Header));
}

PESection::~PESection()
{
  if(this->Data)
    delete Data;
}

PEImport::PEImport()
{
  this->FirstThunk = 0;
  memset(&this->FileName, 0, sizeof(this->FileName));
}

PEImport::~PEImport()
{
  size_t i;
  for(i = 0; i < this->FuncList.size(); i++)
    delete this->FuncList[i];

  this->FuncList.clear();
}

PEImportFunc::PEImportFunc()
{
  this->Hint = 0;
  memset(&this->FuncName, 0, sizeof(this->FuncName));
}

void PE::PadTo(FILE *f, size_t addr)
{
  static BYTE Zeros[256];
  size_t Offset = ftell(f);

  if(addr < Offset)
    return;

  size_t Diff = addr - Offset;

  while(Diff > 0)
  {
    if(Diff > sizeof(Zeros))
    {
      Diff -= sizeof(Zeros);
      fwrite(Zeros, 1, sizeof(Zeros), f);
    }
    else
    {
      fwrite(Zeros, 1, Diff, f);
      Diff = 0;
    }
  }
}

void PE::AlignTo(FILE *f, size_t align)
{
  size_t Offset = ftell(f);
  if(Offset % align == 0)
    return;

  this->PadTo(f, (Offset / align + 1) * align);
}


PE::PE()
{
  memset(&this->DosHeader, 0, sizeof(this->DosHeader));
  memset(&this->NtHeaders, 0, sizeof(this->NtHeaders));
}

PE::~PE()
{
  size_t i;
  for(i = 0; i < this->Sections.size(); i++)
    delete this->Sections[i];

  this->Sections.clear();

  for(i = 0; i < this->Imports.size(); i++)
    delete this->Imports[i];

  this->Imports.clear();
}

WORD PE::RVAtoSection(DWORD rva)
{
  size_t i;
  for(i = 0; i < this->Sections.size(); i++)
  {
    if(this->Sections[i]->Header.VirtualAddress <= rva &&
       this->Sections[i]->Header.SizeOfRawData + this->Sections[i]->Header.VirtualAddress > rva)
      return i;
  }

  return 0xffff;
}

BYTE *PE::RVAtoPtr(DWORD rva)
{
  WORD SecNo = this->RVAtoSection(rva);
  if(SecNo == 0xffff)
    return NULL;

  return &this->Sections[SecNo]->Data[rva - this->Sections[SecNo]->Header.VirtualAddress];
}

bool PE::Load(const char *name)
{
  FILE *f = fopen(name, "rb");
  if(!f)
    return false;

  fread(&this->DosHeader, 1, sizeof(this->DosHeader), f);
  fseek(f, this->DosHeader.e_lfanew, SEEK_SET);
  fread(&this->NtHeaders, 1, sizeof(this->NtHeaders), f);

  // sections
  int i;
  for(i = 0; i < this->NtHeaders.FileHeader.NumberOfSections; i++)
  {
    PESection *Section = new PESection;
    fread(&Section->Header, 1, sizeof(Section->Header), f);
    size_t Offset = ftell(f);
    fseek(f, Section->Header.PointerToRawData, SEEK_SET);
    Section->Data = new BYTE[Section->Header.SizeOfRawData];
    fread(Section->Data, 1, Section->Header.SizeOfRawData, f);
    fseek(f, Offset, SEEK_SET);
    this->Sections.push_back(Section);
  }

  // imports
#if 0
  DWORD ImportDescRVA = this->NtHeaders.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress;
  IMAGE_IMPORT_DESCRIPTOR *ImportDesc = (IMAGE_IMPORT_DESCRIPTOR *)this->RVAtoPtr(ImportDescRVA);
  if(!ImportDesc)
    return false;

  for(; ImportDesc->Name; ImportDesc++)
  {
    PEImport *Import = new PEImport;
    strncpy(Import->FileName, (const char*)this->RVAtoPtr(ImportDesc->Name), 256);
    Import->FirstThunk = ImportDesc->FirstThunk;
    this->Imports.push_back(Import);

    // imported function names
    IMAGE_THUNK_DATA32 *ThunkData = (IMAGE_THUNK_DATA32*)this->RVAtoPtr(ImportDesc->OriginalFirstThunk);
    for(; ThunkData->u1.AddressOfData; ThunkData++)
    {
      if(ThunkData->u1.Ordinal & IMAGE_ORDINAL_FLAG32)
	continue; // not supported
      
      IMAGE_IMPORT_BY_NAME *ByName = (IMAGE_IMPORT_BY_NAME*)this->RVAtoPtr(ThunkData->u1.AddressOfData);
      PEImportFunc *Func = new PEImportFunc;
      Func->Hint = ByName->Hint;
      strncpy(Func->FuncName, (const char*)ByName->Name, 256);
      Import->FuncList.push_back(Func);
    }
  }
#endif
  fclose(f);
  return true;
}

bool PE::Dump(const char *name)
{
  FILE *f = fopen(name, "wb");
  if(!f)
    return false;

  fwrite(&this->DosHeader, 1, sizeof(this->DosHeader), f);
  this->PadTo(f, this->DosHeader.e_lfanew);
  fwrite(&this->NtHeaders, 1, sizeof(this->NtHeaders), f);
  
  size_t Offset = ftell(f);
  this->PadTo(f, Offset + sizeof(IMAGE_SECTION_HEADER) * this->NtHeaders.FileHeader.NumberOfSections);

  int i;
  for(i = 0; i < this->NtHeaders.FileHeader.NumberOfSections; i++)
  {
    this->AlignTo(f, this->NtHeaders.OptionalHeader.FileAlignment);
    this->Sections[i]->Header.PointerToRawData = ftell(f);
    fwrite(this->Sections[i]->Data, 1, this->Sections[i]->Header.SizeOfRawData, f);
  }

  fseek(f, Offset, SEEK_SET);
  for(i = 0; i < this->NtHeaders.FileHeader.NumberOfSections; i++)
    fwrite(&this->Sections[i]->Header, 1, sizeof(this->Sections[i]->Header), f);
 
 
  fclose(f);
  return true;
}


